package ru.landgrafhomyak.bgtu.computer_graphics1

import kotlinx.browser.document
import kotlinx.browser.window
import org.khronos.webgl.Float32Array
import org.khronos.webgl.Int32Array
import org.w3c.dom.HTMLCanvasElement
import ru.landgrafhomyak.bgtu.computer_graphics1.model_builder.Matrix4x4
import ru.landgrafhomyak.bgtu.computer_graphics1.model_builder.Point2F
import ru.landgrafhomyak.bgtu.computer_graphics1.model_builder.Point4F

fun main() {
	val canvas = document.getElementById("canvas")!! as HTMLCanvasElement

	val gl: dynamic = canvas.getContext("webgl2")!! /*as WebGLRenderingContext*/


	val vertexShader = gl.createShader(gl.VERTEX_SHADER);
	gl.shaderSource(vertexShader, vertexShaderSource)
	gl.compileShader(vertexShader)
	println("vertex shader compilation status: ${gl.getShaderParameter(vertexShader, gl.COMPILE_STATUS)}")
	println("vertex shader compilation log: ${gl.getShaderInfoLog(vertexShader)}")

	val fragmentShader = gl.createShader(gl.FRAGMENT_SHADER)
	gl.shaderSource(fragmentShader, fragmentShaderSource)
	gl.compileShader(fragmentShader)
	println("fragment shader compilation status: ${gl.getShaderParameter(fragmentShader, gl.COMPILE_STATUS)}")
	println("fragment shader compilation log: ${gl.getShaderInfoLog(fragmentShader)}")


	val shaderProgram = gl.createProgram()
	gl.attachShader(shaderProgram, vertexShader)
	gl.attachShader(shaderProgram, fragmentShader)
	gl.linkProgram(shaderProgram)
	println("program compilation log: ${gl.getProgramInfoLog(shaderProgram)}")
	gl.useProgram(shaderProgram)
	val transformLocation = gl.getUniformLocation(shaderProgram, "transform")
	val projectionLocation = gl.getUniformLocation(shaderProgram, "projection")
	val screenLocation = gl.getUniformLocation(shaderProgram, "screen_scale")


	val modelVertexesBuffer = gl.createBuffer()
	val textureVertexesBuffer = gl.createBuffer()
	val textureTypesBuffer = gl.createBuffer()

	gl.bindBuffer(gl.ARRAY_BUFFER, modelVertexesBuffer)
	println(model.modelVertexes.joinToString())
	gl.bufferData(
		gl.ARRAY_BUFFER,
		Float32Array(model.modelVertexes.toTypedArray()),
		gl.STATIC_DRAW
	)
	val vpa = gl.getAttribLocation(shaderProgram, "vertex_pos")
	gl.vertexAttribPointer(vpa, 3, gl.FLOAT, false, 0, 0)
	gl.enableVertexAttribArray(vpa)

	gl.bindBuffer(gl.ARRAY_BUFFER, textureVertexesBuffer)
	gl.bufferData(
		gl.ARRAY_BUFFER,
		Float32Array(model.textureVertexes.toTypedArray()),
		gl.STATIC_DRAW
	)
	val tpa = gl.getAttribLocation(shaderProgram, "texture_pos")
	gl.vertexAttribPointer(tpa, 2, gl.FLOAT, false, 0, 0)
	gl.enableVertexAttribArray(tpa)

	gl.bindBuffer(gl.ARRAY_BUFFER, textureTypesBuffer)
	gl.bufferData(
		gl.ARRAY_BUFFER,
		Int32Array(model.textureTypes.map { t -> t.ordinal }.toTypedArray()),
		gl.STATIC_DRAW
	)
	val tta = gl.getAttribLocation(shaderProgram, "texture_type")
	gl.vertexAttribIPointer(tta, 1, gl.INT, false, 0, 0)
	gl.enableVertexAttribArray(tta)

	gl.enable(gl.DEPTH_TEST)

	gl.enable(gl.BLEND);
	gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);

	var xAngle = 0.0f
	var yAngle = 0.0f
	var distance = 0.0f
	val xAngleStep = 0.1f
	val yAngleStep = 0.1f
	val distanceStep = 0.1f


	val redraw = {
		gl.viewport(0, 0, canvas.width, canvas.height);
		gl.clearColor(0f, 0f, 0f, 1f);
		gl.clear(gl.COLOR_BUFFER_BIT or gl.DEPTH_BUFFER_BIT);
		val t = Matrix4x4.shift(0f, 0f, distance) * Matrix4x4.rotateY(yAngle) * Matrix4x4.rotateX(xAngle)
		val po = Matrix4x4.projectionOrtho(-2f, 2f, -2f, 2f, -2f, 2f);
		val pc = Matrix4x4.projectionCentralInfinite(-2f, 2f, -2f, 2f, 1f) * Matrix4x4.shift(0f, 0f, -3.3f) * Matrix4x4.scale(3f, 3f, -1f);
		val s: Point2F
		if (canvas.width > canvas.height) {
			s = Point2F(canvas.height.toFloat() / canvas.width.toFloat(), 1f)
		} else {
			s = Point2F(1f, canvas.width.toFloat() / canvas.height.toFloat())
		}
		gl.uniformMatrix4fv(transformLocation, false, t.toFloatArray())
		gl.uniformMatrix4fv(projectionLocation, false, pc.toFloatArray())
		gl.uniform2f(screenLocation, s.x, s.y)
		gl.drawArrays(gl.TRIANGLES, 0, model.textureTypes.size)
		gl.flush()
	}
	val p = Matrix4x4.projectionCentral(-2f, 2f, -2f, 2f, 2f, 4f);

	println(p * Point4F(-1f, -1f, -2f, 1f))
	println(p * Point4F(-1f, 1f, -3f, 1f))
	println(p * Point4F(1f, 1f, -4f, 1f))

	redraw()

	window.onkeydown = { keyEvent ->
		when (keyEvent.code) {
			"ArrowDown" -> {
				xAngle += xAngleStep;
				redraw()
			}

			"ArrowUp" -> {
				xAngle -= xAngleStep;
				redraw()
			}

			"ArrowLeft" -> {
				yAngle -= yAngleStep;
				redraw()
			}

			"ArrowRight" -> {
				yAngle += yAngleStep;
				redraw()
			}

			"PageUp" -> {
				distance += distanceStep;
				redraw()
			}

			"PageDown" -> {
				distance -= distanceStep;
				if (distance < 0) distance = 0f
				redraw()
			}
		}
	}

	window.onresize = { _ ->
		canvas.width = document.documentElement!!.clientWidth;
		canvas.height = document.documentElement!!.clientHeight;
		redraw()
	}
	window.onload = { _ ->
		canvas.width = document.documentElement!!.clientWidth;
		canvas.height = document.documentElement!!.clientHeight;
		redraw()
	}
}


