package ru.landgrafhomyak.bgtu.computer_graphics1.model_builder

import kotlin.math.cos
import kotlin.math.sin

class Matrix4x4 private constructor(
	private val v00: Float,
	private val v01: Float,
	private val v02: Float,
	private val v03: Float,
	private val v10: Float,
	private val v11: Float,
	private val v12: Float,
	private val v13: Float,
	private val v20: Float,
	private val v21: Float,
	private val v22: Float,
	private val v23: Float,
	private val v30: Float,
	private val v31: Float,
	private val v32: Float,
	private val v33: Float
) {
	private inline fun _foldMatrix(other: Matrix4x4, operator: (rowIndex: Int, columnIndex: Int, self: Float, other: Float) -> Float): Matrix4x4 = Matrix4x4(
		operator(0, 0, this.v00, other.v00), operator(0, 1, this.v01, other.v01), operator(0, 2, this.v02, other.v02), operator(0, 3, this.v03, other.v03),
		operator(1, 0, this.v10, other.v10), operator(1, 1, this.v11, other.v11), operator(1, 2, this.v12, other.v12), operator(1, 3, this.v13, other.v13),
		operator(2, 0, this.v20, other.v20), operator(2, 1, this.v21, other.v21), operator(2, 2, this.v22, other.v22), operator(2, 3, this.v23, other.v23),
		operator(3, 0, this.v30, other.v30), operator(3, 1, this.v31, other.v31), operator(3, 2, this.v32, other.v32), operator(3, 3, this.v33, other.v33),
	)

	operator fun get(row: Int, column: Int) = when (row * 4 + column) {
		0 -> this.v00
		1 -> this.v01
		2 -> this.v02
		3 -> this.v03
		4 -> this.v10
		5 -> this.v11
		6 -> this.v12
		7 -> this.v13
		8 -> this.v20
		9 -> this.v21
		10 -> this.v22
		11 -> this.v23
		12 -> this.v30
		13 -> this.v31
		14 -> this.v32
		15 -> this.v33
		else -> throw IllegalArgumentException()
	}

	operator fun times(other: Matrix4x4): Matrix4x4 =
		this._foldMatrix(other) { y, x, _, _ ->
			return@_foldMatrix this[y, 0] * other[0, x] + this[y, 1] * other[1, x] + this[y, 2] * other[2, x] + this[y, 3] * other[3, x];
		}

	fun toFloatArray() = floatArrayOf(
		this.v00, this.v01, this.v02, this.v03,
		this.v10, this.v11, this.v12, this.v13,
		this.v20, this.v21, this.v22, this.v23,
		this.v30, this.v31, this.v32, this.v33
	)

	fun transposed() = Matrix4x4(
		this.v00, this.v10, this.v20, this.v30,
		this.v01, this.v11, this.v21, this.v31,
		this.v02, this.v12, this.v22, this.v32,
		this.v03, this.v13, this.v23, this.v33,
	)

	operator fun times(p: Point3F) = Point3F(
		x = this.v00 * p.x + this.v01 * p.y + this.v02 * p.z + this.v03,
		y = this.v10 * p.x + this.v11 * p.y + this.v12 * p.z + this.v13,
		z = this.v20 * p.x + this.v21 * p.y + this.v22 * p.z + this.v23,
	)

	operator fun times(p: Point4F) = Point4F(
		x = this.v00 * p.x + this.v01 * p.y + this.v02 * p.z + this.v03,
		y = this.v10 * p.x + this.v11 * p.y + this.v12 * p.z + this.v13,
		z = this.v20 * p.x + this.v21 * p.y + this.v22 * p.z + this.v23,
		w = this.v30 * p.x + this.v31 * p.y + this.v32 * p.z + this.v33,
	)

	companion object {
		val I = Matrix4x4(
			1f, 0f, 0f, 0f,
			0f, 1f, 0f, 0f,
			0f, 0f, 1f, 0f,
			0f, 0f, 0f, 1f
		)

		fun scale(x: Float, y: Float, z: Float) = Matrix4x4(
			x, 0f, 0f, 0f,
			0f, y, 0f, 0f,
			0f, 0f, z, 0f,
			0f, 0f, 0f, 1f
		)

		fun scale(s: Float) = this.scale(s, s, s)

		fun rotateX(rad: Float): Matrix4x4 {
			val c = cos(rad)
			val s = sin(rad)
			return Matrix4x4(
				1f, 0f, 0f, 0f,
				0f, c, -s, 0f,
				0f, s, c, 0f,
				0f, 0f, 0f, 1f
			)
		}

		fun rotateY(rad: Float): Matrix4x4 {
			val c = cos(rad)
			val s = sin(rad)
			return Matrix4x4(
				c, 0f, s, 0f,
				0f, 1f, 0f, 0f,
				-s, 0f, c, 0f,
				0f, 0f, 0f, 1f
			)
		}

		fun rotateZ(rad: Float): Matrix4x4 {
			val c = cos(rad)
			val s = sin(rad)
			return Matrix4x4(
				c, -s, 0f, 0f,
				s, c, 0f, 0f,
				0f, 0f, 1f, 0f,
				0f, 0f, 0f, 1f
			)
		}

		fun shift(x: Float, y: Float, z: Float) = Matrix4x4(
			1f, 0f, 0f, x,
			0f, 1f, 0f, y,
			0f, 0f, 1f, z,
			0f, 0f, 0f, 1f
		)

		fun projectionOrtho(l: Float, r: Float, b: Float, t: Float, n: Float, f: Float) = Matrix4x4(
			2f / (r - l), 0f, 0f, -(r + l) / (r - l),
			0f, 2f / (t - b), 0f, -(t + b) / (t - b),
			0f, 0f, 2f / (f - n), -(f + n) / (f - n),
			0f, 0f, 0f, 1f
		)

		fun projectionCentral(l: Float, r: Float, b: Float, t: Float, n: Float, f: Float) = Matrix4x4(
			2f * n / (r - l), 0f, (r + l) / (r - l), 0f,
			0f, 2f * n / (t - b), (t + b) / (t - b), 0f,
			0f, 0f, -(f + n) / (f - n), -2f * n * f / (f - n),
			0f, 0f, -1f, 0f
		)

		fun projectionCentralInfinite(l: Float, r: Float, b: Float, t: Float, n: Float) = Matrix4x4(
			2f * n / (r - l), 0f, (r + l) / (r - l), 0f,
			0f, 2f * n / (t - b), (t + b) / (t - b), 0f,
			0f, 0f, -1f, -2f * n,
			0f, 0f, -1f, 0f
		)
	}

}